r"""Rand index (`randindex`)"""

from ruptures.metrics.sanity_check import sanity_check


def randindex(bkps1, bkps2):
    """Computes the Rand index (between 0 and 1) between two segmentations.

    The Rand index (RI) measures the similarity between two segmentations and
    is equal to the proportion of aggreement between two partitions.

    RI is between 0 (total disagreement) and 1 (total agreement).
    This function uses the efficient implementation of [1].

    [1] Prates, L. (2021). A more efficient algorithm to compute the Rand Index for
    change-point problems. ArXiv:2112.03738.

    Args:
        bkps1 (list): sorted list of the last index of each regime.
        bkps2 (list): sorted list of the last index of each regime.

    Returns:
        float: Rand index
    """
    sanity_check(bkps1, bkps2)
    n_samples = bkps1[-1]
    bkps1_with_0 = [0] + bkps1
    bkps2_with_0 = [0] + bkps2
    n_bkps1 = len(bkps1)
    n_bkps2 = len(bkps2)

    disagreement = 0
    beginj: int = 0  # avoids unnecessary computations
    for index_bkps1 in range(n_bkps1):
        start1: int = bkps1_with_0[index_bkps1]
        end1: int = bkps1_with_0[index_bkps1 + 1]
        for index_bkps2 in range(beginj, n_bkps2):
            start2: int = bkps2_with_0[index_bkps2]
            end2: int = bkps2_with_0[index_bkps2 + 1]
            nij = max(min(end1, end2) - max(start1, start2), 0)
            disagreement += nij * abs(end1 - end2)

            # we can skip the rest of the iteration, nij will be 0
            if end1 < end2:
                break
            else:
                beginj = index_bkps2 + 1

    disagreement /= n_samples * (n_samples - 1) / 2
    return 1.0 - disagreement



import numpy as np
from ruptures.utils import Bnode


def test_bnode():
    left = Bnode(start=100, end=120, val=1)
    right = Bnode(start=120, end=200, val=1)

    # bad merging, no right leaf
    merged_node = Bnode(start=left.start, end=right.end, left=left, right=None, val=3)
    assert merged_node.gain == 0

    # bad merging, no left leaf
    merged_node = Bnode(start=left.start, end=right.end, left=None, right=right, val=3)
    assert merged_node.gain == 0

    # bad merging, negative infinit val
    merged_node = Bnode(
        start=left.start, end=right.end, left=left, right=right, val=-np.inf
    )
    assert merged_node.gain == 0

    # normal merging
    merged_node = Bnode(
        start=left.start,
        end=right.end,
        left=left,
        right=right,
        val=left.val + right.val + 1,
    )
    assert merged_node.gain == merged_node.val - (left.val + right.val)



"""Hamming metric for segmentation."""

from ruptures.metrics.randindex import randindex


def hamming(bkps1, bkps2):
    """Modified Hamming distance for partitions. For all pair of points (x, y),
    x != y, the functions computes the number of times the two partitions
    disagree. The result is scaled to be within 0 and 1.

    Args:
        bkps1 (list): list of the last index of each regime.
        bkps2 (list): list of the last index of each regime.

    Returns:
        float: Hamming distance.
    """
    return 1 - randindex(bkps1=bkps1, bkps2=bkps2)



r"""Shift in linear model."""

import numpy as np

from . import pw_constant


def pw_linear(n_samples=200, n_features=1, n_bkps=3, noise_std=None, seed=None):
    """Return piecewise linear signal and the associated changepoints.

    Args:
        n_samples (int, optional): signal length
        n_features (int, optional): number of covariates
        n_bkps (int, optional): number of change points
        noise_std (float, optional): noise std. If None, no noise is added
        seed (int): random seed
    Returns:
        tuple: signal of shape (n_samples, n_features+1), list of breakpoints
    """
    rng = np.random.default_rng(seed=seed)
    covar = rng.normal(size=(n_samples, n_features))
    linear_coeff, bkps = pw_constant(
        n_samples=n_samples,
        n_bkps=n_bkps,
        n_features=n_features,
        noise_std=None,
        seed=seed,
    )
    var = np.sum(linear_coeff * covar, axis=1)
    if noise_std is not None:
        var += rng.normal(scale=noise_std, size=var.shape)
    signal = np.c_[var, covar]
    return signal, bkps



r"""Efficient kernel change point detection (dynamic programming)"""

from ruptures.base import BaseEstimator
from ruptures.costs import cost_factory
from ruptures.utils import from_path_matrix_to_bkps_list, sanity_check
from ruptures.exceptions import BadSegmentationParameters
import numpy as np

from ._detection.ekcpd import (
    ekcpd_cosine,
    ekcpd_Gaussian,
    ekcpd_L2,
    ekcpd_pelt_cosine,
    ekcpd_pelt_Gaussian,
    ekcpd_pelt_L2,
)


class KernelCPD(BaseEstimator):
    """Find optimal change points (using dynamic programming or pelt) for the
    special case where the cost function derives from a kernel function.

    Given a segment model, it computes the best partition for which the
    sum of errors is minimum.

    See the [user guide](../../../user-guide/detection/kernelcpd) for
    more information.
    """

    def __init__(self, kernel="linear", min_size=2, jump=1, params=None):
        r"""Creates a KernelCPD instance.

        Available kernels:

        - `linear`: $k(x,y) = x^T y$.
        - `rbf`: $k(x, y) = exp(\gamma \|x-y\|^2)$ where $\gamma>0$
        (`gamma`) is a user-defined parameter.
        - `cosine`: $k(x,y)= (x^T y)/(\|x\|\|y\|)$.

        Args:
            kernel (str, optional): name of the kernel, ["linear", "rbf", "cosine"]
            min_size (int, optional): minimum segment length.
            jump (int, optional): not considered, set to 1.
            params (dict, optional): a dictionary of parameters for the kernel instance

        Raises:
            AssertionError: if the kernel is not implemented.
        """
        self.kernel_name = kernel
        err_msg = "Kernel not found: {}.".format(self.kernel_name)
        assert self.kernel_name in ["linear", "rbf", "cosine"], err_msg
        self.model_name = "l2" if self.kernel_name == "linear" else self.kernel_name
        self.params = params
        # load the associated cost function
        if self.params is None:
            self.cost = cost_factory(model=self.model_name)
        else:
            self.cost = cost_factory(model=self.model_name, **self.params)
        self.min_size = max(min_size, self.cost.min_size)

        self.jump = 1  # set to 1
        self.n_samples = None
        self.segmentations_dict = dict()  # {n_bkps: bkps_list}

    def fit(self, signal) -> "KernelCPD":
        """Update some parameters (no computation in this function).

        Args:
            signal (array): signal. Shape (n_samples, n_features) or (n_samples,).

        Returns:
            self
        """
        # update some params
        self.segmentations_dict = dict()
        self.cost.fit(signal.astype(np.double))
        self.n_samples = signal.shape[0]
        return self

    def predict(self, n_bkps=None, pen=None):
        """Return the optimal breakpoints. Must be called after the fit method.

        The breakpoints are associated with the signal passed to
        [`fit()`][ruptures.detection.kernelcpd.KernelCPD.fit].

        Args:
            n_bkps (int, optional): Number of change points. Defaults to None.
            pen (float, optional): penalty value (>0). Defaults to None. Not considered
                if n_bkps is not None.

        Raises:
            AssertionError: if `pen` or `n_bkps` is not strictly positive.
            BadSegmentationParameters: in case of impossible segmentation
                configuration

        Returns:
            list[int]: sorted list of breakpoints
        """
        # Our KernelCPD implementation with Pelt implies that we have at least one change point
        # raise an exception in case of impossible segmentation configuration
        if not sanity_check(
            n_samples=self.cost.signal.shape[0],
            n_bkps=1 if n_bkps is None else n_bkps,
            jump=self.jump,
            min_size=self.min_size,
        ):
            raise BadSegmentationParameters

        # dynamic programming if the user passed a number change points
        if n_bkps is not None:
            n_bkps = int(n_bkps)
            err_msg = "The number of changes must be positive: {}".format(n_bkps)
            assert n_bkps > 0, err_msg
            # if we have already computed it, return it without computations.
            if n_bkps in self.segmentations_dict:
                return self.segmentations_dict[n_bkps]
            # otherwise, call the C function
            if self.kernel_name == "linear":
                path_matrix_flat = ekcpd_L2(self.cost.signal, n_bkps, self.min_size)
            elif self.kernel_name == "rbf":
                path_matrix_flat = ekcpd_Gaussian(
                    self.cost.signal, n_bkps, self.min_size, self.cost.gamma
                )
            elif self.kernel_name == "cosine":
                path_matrix_flat = ekcpd_cosine(self.cost.signal, n_bkps, self.min_size)
            # from the path matrix, get all segmentation for k=1,...,n_bkps changes
            for k in range(1, n_bkps + 1):
                self.segmentations_dict[k] = from_path_matrix_to_bkps_list(
                    path_matrix_flat, k, self.n_samples, n_bkps, self.jump
                )
            return self.segmentations_dict[n_bkps]

        # Call pelt if the user passed a penalty
        if pen is not None:
            assert pen > 0, "The penalty must be positive: {}".format(pen)
            if self.kernel_name == "linear":
                path_matrix = ekcpd_pelt_L2(self.cost.signal, pen, self.min_size)
            elif self.kernel_name == "rbf":
                path_matrix = ekcpd_pelt_Gaussian(
                    self.cost.signal, pen, self.min_size, self.cost.gamma
                )
            elif self.kernel_name == "cosine":
                path_matrix = ekcpd_pelt_cosine(self.cost.signal, pen, self.min_size)

            my_bkps = list()
            ind = self.n_samples
            while ind > 0:
                my_bkps.append(ind)
                ind = path_matrix[ind]
            return my_bkps[::-1]

    def fit_predict(self, signal, n_bkps=None, pen=None):
        """Fit to the signal and return the optimal breakpoints.

        Helper method to call fit and predict once

        Args:
            signal (array): signal. Shape (n_samples, n_features) or (n_samples,).
            n_bkps (int, optional): Number of change points. Defaults to None.
            pen (float, optional): penalty value (>0). Defaults to None. Not considered if n_bkps is not None.

        Returns:
            list: sorted list of breakpoints
        """
        self.fit(signal)
        return self.predict(n_bkps=n_bkps, pen=pen)



r"""Precision and recall."""

from itertools import product

from ruptures.metrics.sanity_check import sanity_check


def precision_recall(true_bkps, my_bkps, margin=10):
    """Calculate the precision/recall of an estimated segmentation compared
    with the true segmentation.

    Args:
        true_bkps (list): list of the last index of each regime (true
            partition).
        my_bkps (list): list of the last index of each regime (computed
            partition).
        margin (int, optional): allowed error (in points).

    Returns:
        tuple: (precision, recall)
    """
    sanity_check(true_bkps, my_bkps)
    assert margin > 0, "Margin of error must be positive (margin = {})".format(margin)

    if len(my_bkps) == 1:
        return 0, 0

    used = set()
    true_pos = set(
        true_b
        for true_b, my_b in product(true_bkps[:-1], my_bkps[:-1])
        if my_b - margin < true_b < my_b + margin
        and not (my_b in used or used.add(my_b))
    )

    tp_ = len(true_pos)
    precision = tp_ / (len(my_bkps) - 1)
    recall = tp_ / (len(true_bkps) - 1)
    return precision, recall



r"""Hausdorff metric."""

import numpy as np
from scipy.spatial.distance import cdist
from ruptures.metrics.sanity_check import sanity_check


def hausdorff(bkps1, bkps2):
    """Compute the Hausdorff distance between changepoints.

    Args:
        bkps1 (list): list of the last index of each regime.
        bkps2 (list): list of the last index of each regime.

    Returns:
        float: Hausdorff distance.
    """
    sanity_check(bkps1, bkps2)
    bkps1_arr = np.array(bkps1[:-1]).reshape(-1, 1)
    bkps2_arr = np.array(bkps2[:-1]).reshape(-1, 1)
    pw_dist = cdist(bkps1_arr, bkps2_arr)
    res = max(pw_dist.min(axis=0).max(), pw_dist.min(axis=1).max())
    return res



r"""Pelt."""

from math import floor

from ruptures.costs import cost_factory
from ruptures.base import BaseCost, BaseEstimator
from ruptures.exceptions import BadSegmentationParameters
from ruptures.utils import sanity_check


class Pelt(BaseEstimator):
    """Penalized change point detection.

    For a given model and penalty level, computes the segmentation which
    minimizes the constrained sum of approximation errors.
    """

    def __init__(self, model="l2", custom_cost=None, min_size=2, jump=5, params=None):
        """Initialize a Pelt instance.

        Args:
            model (str, optional): segment model, ["l1", "l2", "rbf"]. Not used if ``'custom_cost'`` is not None.
            custom_cost (BaseCost, optional): custom cost function. Defaults to None.
            min_size (int, optional): minimum segment length.
            jump (int, optional): subsample (one every *jump* points).
            params (dict, optional): a dictionary of parameters for the cost instance.
        """
        if custom_cost is not None and isinstance(custom_cost, BaseCost):
            self.cost = custom_cost
        else:
            if params is None:
                self.cost = cost_factory(model=model)
            else:
                self.cost = cost_factory(model=model, **params)
        self.min_size = max(min_size, self.cost.min_size)
        self.jump = jump
        self.n_samples = None

    def _seg(self, pen):
        """Computes the segmentation for a given penalty using PELT (or a list
        of penalties).

        Args:
            penalty (float): penalty value

        Returns:
            dict: partition dict {(start, end): cost value,...}
        """
        # initialization
        # partitions[t] contains the optimal partition of signal[0:t]
        partitions = dict()  # this dict will be recursively filled
        partitions[0] = {(0, 0): 0}
        admissible = []

        # Recursion
        ind = [k for k in range(0, self.n_samples, self.jump) if k >= self.min_size]
        ind += [self.n_samples]
        for bkp in ind:
            # adding a point to the admissible set from the previous loop.
            new_adm_pt = floor((bkp - self.min_size) / self.jump)
            new_adm_pt *= self.jump
            admissible.append(new_adm_pt)

            subproblems = list()
            for t in admissible:
                # left partition
                try:
                    tmp_partition = partitions[t].copy()
                except KeyError:  # no partition of 0:t exists
                    continue
                # we update with the right partition
                tmp_partition.update({(t, bkp): self.cost.error(t, bkp) + pen})
                subproblems.append(tmp_partition)

            # finding the optimal partition
            partitions[bkp] = min(subproblems, key=lambda d: sum(d.values()))
            # trimming the admissible set
            admissible = [
                t
                for t, partition in zip(admissible, subproblems)
                if sum(partition.values()) <= sum(partitions[bkp].values()) + pen
            ]

        best_partition = partitions[self.n_samples]
        del best_partition[(0, 0)]
        return best_partition

    def fit(self, signal) -> "Pelt":
        """Set params.

        Args:
            signal (array): signal to segment. Shape (n_samples, n_features) or (n_samples,).

        Returns:
            self
        """
        # update params
        self.cost.fit(signal)
        if signal.ndim == 1:
            (n_samples,) = signal.shape
        else:
            n_samples, _ = signal.shape
        self.n_samples = n_samples
        return self

    def predict(self, pen):
        """Return the optimal breakpoints.

        Must be called after the fit method. The breakpoints are associated with the signal passed
        to [`fit()`][ruptures.detection.pelt.Pelt.fit].

        Args:
            pen (float): penalty value (>0)

        Raises:
            BadSegmentationParameters: in case of impossible segmentation
                configuration

        Returns:
            list: sorted list of breakpoints
        """
        # raise an exception in case of impossible segmentation configuration
        if not sanity_check(
            n_samples=self.cost.signal.shape[0],
            n_bkps=0,
            jump=self.jump,
            min_size=self.min_size,
        ):
            raise BadSegmentationParameters

        partition = self._seg(pen)
        bkps = sorted(e for s, e in partition.keys())
        return bkps

    def fit_predict(self, signal, pen):
        """Fit to the signal and return the optimal breakpoints.

        Helper method to call fit and predict once

        Args:
            signal (array): signal. Shape (n_samples, n_features) or (n_samples,).
            pen (float): penalty value (>0)

        Returns:
            list: sorted list of breakpoints
        """
        self.fit(signal)
        return self.predict(pen)



r"""Draw a random partition."""

import numpy as np


def draw_bkps(n_samples=100, n_bkps=3, seed=None):
    """Draw a random partition with specified number of samples and specified
    number of changes."""
    rng = np.random.default_rng(seed=seed)
    alpha = np.ones(n_bkps + 1) / (n_bkps + 1) * 2000
    bkps = np.cumsum(rng.dirichlet(alpha) * n_samples).astype(int).tolist()
    bkps[-1] = n_samples
    return bkps



from copy import deepcopy
from itertools import product

import numpy as np
import pytest

from ruptures.costs import CostAR
from ruptures.datasets import pw_constant
from ruptures.detection import Binseg, BottomUp, Dynp, Pelt, Window, KernelCPD
from ruptures.exceptions import BadSegmentationParameters


@pytest.fixture(scope="module")
def signal_bkps_5D_n10():
    signal, bkps = pw_constant(n_samples=10, n_features=5, noise_std=1)
    return signal, bkps


@pytest.fixture(scope="module")
def signal_bkps_5D():
    signal, bkps = pw_constant(n_features=5, noise_std=1)
    return signal, bkps


@pytest.fixture(scope="module")
def signal_bkps_1D():
    signal, bkps = pw_constant(noise_std=1)
    return signal.astype(np.float32), bkps


@pytest.fixture(scope="module")
def signal_bkps_5D_no_noise():
    signal, bkps = pw_constant(n_features=5, noise_std=0)
    return signal, bkps


@pytest.fixture(scope="module")
def signal_bkps_1D_no_noise():
    signal, bkps = pw_constant(noise_std=0)
    return signal, bkps


@pytest.fixture(scope="module")
def signal_bkps_1D_constant():
    signal, bkps = np.zeros(200), [200]
    return signal, bkps


@pytest.mark.parametrize("algo", [Binseg, BottomUp, Dynp, Pelt, Window])
def test_empty(signal_bkps_1D, algo):
    signal, _ = signal_bkps_1D
    algo().fit(signal).predict(1)
    algo().fit_predict(signal, 1)


@pytest.mark.parametrize(
    "algo, model",
    product(
        [Binseg, BottomUp, Window],
        ["l1", "l2", "ar", "normal", "rbf", "rank", "mahalanobis"],
    ),
)
def test_model_1D(signal_bkps_1D, algo, model):
    signal, _ = signal_bkps_1D
    algo(model=model).fit_predict(signal, pen=1)
    ret = algo(model=model).fit_predict(signal, n_bkps=1)
    assert len(ret) == 2
    assert ret[-1] == signal.shape[0]
    algo(model=model).fit_predict(signal, epsilon=10)


@pytest.mark.parametrize(
    "algo, model",
    product([Dynp, Pelt], ["l1", "l2", "ar", "normal", "rbf", "rank", "mahalanobis"]),
)
def test_model_1D_bis(signal_bkps_1D, algo, model):
    signal, _ = signal_bkps_1D
    algo_t = algo(model=model)
    ret = algo_t.fit_predict(signal, 1)
    if isinstance(algo_t, Dynp):
        assert len(ret) == 2
    assert ret[-1] == signal.shape[0]


@pytest.mark.parametrize(
    "algo, model",
    product(
        [Dynp, Binseg, BottomUp, Window, Pelt],
        ["l1", "l2", "ar", "normal", "rbf", "rank"],
    ),
)
def test_model_1D_constant(signal_bkps_1D_constant, algo, model):
    signal, _ = signal_bkps_1D_constant
    algo = algo(model=model)
    if isinstance(algo, Dynp) or isinstance(algo, BottomUp) or isinstance(algo, Binseg):
        ret = algo.fit_predict(signal=signal, n_bkps=1)
        # Even with constant signals, return the specified number of
        # change-points.
        assert len(ret) == 2
    if isinstance(algo, Window):
        ret = algo.fit_predict(signal=signal, n_bkps=1)
        # With constant signal, this search method returns 0 change-point.
        assert len(ret) == 1
    if isinstance(algo, Pelt):
        ret = algo.fit_predict(signal=signal, pen=1)
        # With constant signal, this search method returns 0 change-point.
        assert len(ret) == 1
    assert ret[-1] == signal.shape[0], "The last change-point is equal to"
    " n_samples."


@pytest.mark.parametrize("algo", [Binseg, Window])
def test_costnormal_on_constant_old_behaviour(signal_bkps_1D_constant, algo):
    signal, _ = signal_bkps_1D_constant
    algo = algo(model="normal", params={"add_small_diag": False})
    ret = algo.fit_predict(signal=signal, n_bkps=2)
    # With constant signal, this search method returns 0 change-point.
    assert len(ret) == 1
    # The last change-point is equal to n_samples.
    assert ret[-1] == signal.shape[0], "The last change-point is equal to"
    " n_samples."


@pytest.mark.parametrize(
    "algo, model",
    product(
        [Binseg, BottomUp, Window],
        ["l1", "l2", "linear", "normal", "rbf", "rank", "mahalanobis"],
    ),
)
def test_model_5D(signal_bkps_5D, algo, model):
    signal, _ = signal_bkps_5D
    algo(model=model).fit_predict(signal, pen=1)
    ret = algo(model=model).fit_predict(signal, n_bkps=1)
    assert len(ret) == 2
    algo(model=model).fit_predict(signal, epsilon=10)


@pytest.mark.parametrize(
    "algo, model",
    product(
        [Dynp, Pelt],
        ["l1", "l2", "linear", "normal", "rbf", "rank", "mahalanobis"],
    ),
)
def test_model_5D_bis(signal_bkps_5D, algo, model):
    signal, _ = signal_bkps_5D
    algo_t = algo(model=model)
    ret = algo_t.fit_predict(signal, 1)
    if isinstance(algo_t, Dynp):
        assert len(ret) == 2


@pytest.mark.parametrize("algo", [Binseg, BottomUp, Window, Dynp, Pelt])
def test_custom_cost(signal_bkps_1D, algo):
    signal, _ = signal_bkps_1D
    c = CostAR(order=10)
    algo_t = algo(custom_cost=c)
    ret = algo_t.fit_predict(signal, 1)
    if isinstance(algo_t, Pelt):
        assert len(ret) >= 2
    else:
        assert len(ret) == 2


@pytest.mark.parametrize("algo", [Binseg, BottomUp, Window, Dynp, Pelt])
def test_pass_param_to_cost(signal_bkps_1D, algo):
    signal, _ = signal_bkps_1D
    algo_t = algo(model="ar", params={"order": 10})
    ret = algo_t.fit_predict(signal, 1)
    if isinstance(algo_t, Pelt):
        assert len(ret) >= 2
    else:
        assert len(ret) == 2


@pytest.mark.parametrize(
    "kernel, min_size",
    product(["linear"], [2, 5]),
)
def test_kernelcpd_1D_linear(signal_bkps_1D, kernel, min_size):
    signal, bkps = signal_bkps_1D
    ret = (
        KernelCPD(kernel=kernel, min_size=min_size, jump=1)
        .fit(signal)
        .predict(n_bkps=len(bkps) - 1)
    )
    assert len(ret) == len(bkps)


@pytest.mark.parametrize(
    "kernel, min_size",
    product(["linear"], [2, 5]),
)
def test_kernelcpd_5D_linear(signal_bkps_5D, kernel, min_size):
    signal, bkps = signal_bkps_5D
    ret = (
        KernelCPD(kernel=kernel, min_size=min_size, jump=1)
        .fit(signal)
        .predict(n_bkps=len(bkps) - 1)
    )
    assert len(ret) == len(bkps)


@pytest.mark.parametrize(
    "kernel, min_size",
    product(["rbf"], [2, 5]),
)
def test_kernelcpd_1D_rbf(signal_bkps_1D, kernel, min_size):
    signal, bkps = signal_bkps_1D
    ret = (
        KernelCPD(kernel=kernel, min_size=min_size, jump=1, params={"gamma": 1.5})
        .fit(signal)
        .predict(n_bkps=len(bkps) - 1)
    )
    assert len(ret) == len(bkps)


@pytest.mark.parametrize(
    "kernel, min_size",
    product(["rbf"], [2, 5]),
)
def test_kernelcpd_5D_rbf(signal_bkps_5D, kernel, min_size):
    signal, bkps = signal_bkps_5D
    ret = (
        KernelCPD(kernel=kernel, min_size=min_size, jump=1, params={"gamma": 1.5})
        .fit(signal)
        .predict(n_bkps=len(bkps) - 1)
    )
    assert len(ret) == len(bkps)


@pytest.mark.parametrize(
    "kernel, min_size",
    product(["linear"], [2, 5]),
)
def test_kernelcpd_1D_no_noise_linear(signal_bkps_1D_no_noise, kernel, min_size):
    signal, bkps = signal_bkps_1D_no_noise
    res = (
        KernelCPD(kernel=kernel, min_size=min_size, jump=1)
        .fit(signal)
        .predict(n_bkps=len(bkps) - 1)
    )
    assert res == bkps


@pytest.mark.parametrize(
    "kernel, min_size",
    product(["linear"], [2, 5]),
)
def test_kernelcpd_5D_no_noise_linear(signal_bkps_5D_no_noise, kernel, min_size):
    signal, bkps = signal_bkps_5D_no_noise
    res = (
        KernelCPD(kernel=kernel, min_size=min_size, jump=1)
        .fit(signal)
        .predict(n_bkps=len(bkps) - 1)
    )
    assert res == bkps


@pytest.mark.parametrize(
    "kernel, min_size",
    product(["rbf"], [2, 5]),
)
def test_kernelcpd_1D_no_noise_rbf(signal_bkps_1D_no_noise, kernel, min_size):
    signal, bkps = signal_bkps_1D_no_noise
    res = (
        KernelCPD(kernel=kernel, min_size=min_size, jump=1, params={"gamma": 1.5})
        .fit(signal)
        .predict(n_bkps=len(bkps) - 1)
    )
    assert res == bkps


@pytest.mark.parametrize(
    "kernel, min_size",
    product(["rbf"], [2, 5]),
)
def test_kernelcpd_5D_no_noise_rbf(signal_bkps_5D_no_noise, kernel, min_size):
    signal, bkps = signal_bkps_5D_no_noise
    res = (
        KernelCPD(kernel=kernel, min_size=min_size, jump=1, params={"gamma": 1.5})
        .fit(signal)
        .predict(n_bkps=len(bkps) - 1)
    )
    assert res == bkps


# Exhaustive test of KernelCPD
@pytest.mark.parametrize("kernel", ["linear", "rbf", "cosine"])
def test_kernelcpd(signal_bkps_5D, kernel):
    signal, bkps = signal_bkps_5D
    # Test we do not compute if intermediary results exist
    algo_temp = KernelCPD(kernel=kernel)
    algo_temp.fit(signal).predict(n_bkps=len(bkps) - 1)
    algo_temp.predict(n_bkps=1)
    # Test penalized version
    KernelCPD(kernel=kernel).fit(signal).predict(pen=0.2)
    # Test fit_predict
    KernelCPD(kernel=kernel).fit_predict(signal, pen=0.2)


@pytest.mark.parametrize("kernel", ["linear", "rbf", "cosine"])
def test_kernelcpd_small_signal(signal_bkps_5D_n10, kernel):
    signal, _ = signal_bkps_5D_n10
    algo_temp = KernelCPD(kernel=kernel)
    with pytest.raises(BadSegmentationParameters):
        KernelCPD(kernel=kernel, min_size=10, jump=2).fit_predict(signal, n_bkps=2)
    with pytest.raises(AssertionError):
        KernelCPD(kernel=kernel, min_size=10, jump=2).fit_predict(signal, n_bkps=0)
    with pytest.raises(BadSegmentationParameters):
        KernelCPD(kernel=kernel, min_size=10, jump=2).fit_predict(signal, pen=0.2)
    assert (
        len(KernelCPD(kernel=kernel, min_size=5, jump=2).fit_predict(signal, pen=0.2))
        > 0
    )


@pytest.mark.parametrize("kernel", ["linear", "rbf", "cosine"])
def test_kernelcpd_small_signal_same_result(signal_bkps_5D_n10, kernel):
    signal, _ = signal_bkps_5D_n10
    algo = KernelCPD(kernel=kernel)
    list_of_segmentations = list()
    n_iter = 100
    for _ in range(n_iter):
        bkps = algo.fit(signal=signal).predict(pen=1.0)
        list_of_segmentations.append(bkps)

    # test if all segmentations are equal
    first_bkps = list_of_segmentations[0]
    all_elements_are_equal = all(
        first_bkps == other_bkps for other_bkps in list_of_segmentations[1:]
    )
    err_msg = "KernelCPD returns different segmentations on the same signal."
    assert all_elements_are_equal, err_msg


@pytest.mark.parametrize(
    "algo, model",
    product(
        [Binseg, BottomUp, Window],
        ["l1", "l2", "ar", "normal", "rbf", "rank", "mahalanobis"],
    ),
)
def test_model_small_signal(signal_bkps_5D_n10, algo, model):
    signal, _ = signal_bkps_5D_n10
    with pytest.raises(BadSegmentationParameters):
        algo(model=model, min_size=5, jump=2).fit_predict(signal, n_bkps=2)
    assert len(algo(model=model, min_size=5, jump=2).fit_predict(signal, pen=10**6)) > 0
    assert (
        len(algo(model=model, min_size=5, jump=2).fit_predict(signal, epsilon=10)) > 0
    )
    assert len(algo(model=model, min_size=9, jump=2).fit_predict(signal, pen=10**6)) > 0


@pytest.mark.parametrize(
    "model",
    ["l1", "l2", "ar", "normal", "rbf", "rank", "mahalanobis"],
)
def test_binseg_min_size(signal_bkps_5D_n10, model):
    signal, _ = signal_bkps_5D_n10

    c_bkps = Binseg(model=model, min_size=5, jump=1).fit_predict(signal, n_bkps=1)
    assert all([a == b for a, b in zip(c_bkps, [5, 10])])


@pytest.mark.parametrize(
    "model", ["l1", "l2", "ar", "normal", "rbf", "rank", "mahalanobis"]
)
def test_model_small_signal_dynp(signal_bkps_5D_n10, model):
    signal, _ = signal_bkps_5D_n10
    with pytest.raises(BadSegmentationParameters):
        Dynp(model=model, min_size=5, jump=2).fit_predict(signal, 2)
    with pytest.raises(BadSegmentationParameters):
        Dynp(model=model, min_size=9, jump=2).fit_predict(signal, 2)
    with pytest.raises(BadSegmentationParameters):
        Dynp(model=model, min_size=11, jump=2).fit_predict(signal, 2)
    # Test if it can find the single eligible break point compatible with min_size
    c_bkps = Dynp(model=model, min_size=5, jump=1).fit_predict(signal, 1)
    assert all([a == b for a, b in zip(c_bkps, [5, 10])])


@pytest.mark.parametrize(
    "model", ["l1", "l2", "ar", "normal", "rbf", "rank", "mahalanobis"]
)
def test_model_small_signal_pelt(signal_bkps_5D_n10, model):
    signal, _ = signal_bkps_5D_n10
    with pytest.raises(BadSegmentationParameters):
        Pelt(model=model, min_size=11, jump=2).fit_predict(signal, 2)
    assert len(Pelt(model=model, min_size=10, jump=2).fit_predict(signal, 1.0)) > 0


def test_binseg_deepcopy():
    binseg = Binseg()
    binseg_copy = deepcopy(binseg)
    assert id(binseg.single_bkp) != id(binseg_copy.single_bkp)


def test_min_size_kernelcpd(signal_bkps_1D):
    """Make sure that KernelCPD can handled min_size=1.

    see https://github.com/deepcharles/ruptures/issues/269
    """
    signal, _ = signal_bkps_1D
    # test if the attribute min_size is set to 1
    algo = KernelCPD(kernel="rbf", min_size=1).fit(signal=signal)
    assert algo.min_size == 1

    # test if it can predict segments of size 1
    n_samples = signal.shape[0]
    algo.predict(n_bkps=n_samples - 1)



"""2D piecewise Gaussian process (pw_normal)"""

from itertools import cycle

import numpy as np

from ruptures.utils import draw_bkps


def pw_normal(n_samples=200, n_bkps=3, seed=None):
    """Return a 2D piecewise Gaussian signal and the associated changepoints.

    Args:
        n_samples (int, optional): signal length
        n_bkps (int, optional): number of change points
        seed (int): random seed

    Returns:
        tuple: signal of shape (n_samples, 2), list of breakpoints
    """
    # breakpoints
    bkps = draw_bkps(n_samples, n_bkps, seed=seed)
    # we create the signal
    signal = np.zeros((n_samples, 2), dtype=float)
    cov1 = np.array([[1, 0.9], [0.9, 1]])
    cov2 = np.array([[1, -0.9], [-0.9, 1]])
    rng = np.random.default_rng(seed=seed)
    for sub, cov in zip(np.split(signal, bkps), cycle((cov1, cov2))):
        n_sub, _ = sub.shape
        sub += rng.multivariate_normal([0, 0], cov, size=n_sub)

    return signal, bkps



r"""CostL1 (least absolute deviation)"""

import numpy as np

from ruptures.base import BaseCost
from ruptures.costs import NotEnoughPoints


class CostL1(BaseCost):
    r"""Least absolute deviation."""

    model = "l1"

    def __init__(self) -> None:
        """Initialize the object."""
        self.signal = None
        self.min_size = 2

    def fit(self, signal) -> "CostL1":
        """Set parameters of the instance.

        Args:
            signal (array): signal. Shape (n_samples,) or (n_samples, n_features)

        Returns:
            self
        """
        if signal.ndim == 1:
            self.signal = signal.reshape(-1, 1)
        else:
            self.signal = signal

        return self

    def error(self, start, end) -> float:
        """Return the approximation cost on the segment [start:end].

        Args:
            start (int): start of the segment
            end (int): end of the segment

        Returns:
            segment cost

        Raises:
            NotEnoughPoints: when the segment is too short (less than `min_size` samples).
        """
        if end - start < self.min_size:
            raise NotEnoughPoints
        sub = self.signal[start:end]
        med = np.median(sub, axis=0)

        return abs(sub - med).sum()



import numpy as np
import pytest
from ruptures import Binseg
from ruptures.costs import CostLinear, CostNormal, cost_factory
from ruptures.costs.costml import CostMl
from ruptures.datasets import pw_constant
from ruptures.exceptions import NotEnoughPoints


@pytest.fixture(scope="module")
def signal_bkps_1D():
    signal, bkps = pw_constant(n_features=1, seed=1234567890)
    return signal, bkps


@pytest.fixture(scope="module")
def signal_bkps_1D_noisy():
    signal, bkps = pw_constant(n_features=1, noise_std=1, seed=1234567890)
    return signal, bkps


@pytest.fixture(scope="module")
def signal_bkps_5D():
    signal, bkps = pw_constant(n_features=5, seed=1234567890)
    return signal, bkps


@pytest.fixture(scope="module")
def signal_bkps_5D_noisy():
    signal, bkps = pw_constant(n_features=5, noise_std=1, seed=1234567890)
    return signal, bkps


cost_names = {
    "ar",
    "l1",
    "l2",
    "normal",
    "rbf",
    "rank",
    "clinear",
    "mahalanobis",
    "cosine",
}


@pytest.mark.parametrize("cost_name", cost_names)
def test_costs_1D_names(signal_bkps_1D, cost_name):
    signal, bkps = signal_bkps_1D
    cost = cost_factory(cost_name)
    cost.fit(signal)
    cost.fit(signal.flatten())
    cost.error(0, 100)
    cost.error(100, signal.shape[0])
    cost.error(10, 50)
    cost.sum_of_costs(bkps)
    with pytest.raises(NotEnoughPoints):
        if cost_name == "cosine":
            cost.min_size = 4
            cost.error(1, 2)
        elif cost_name == "l2":
            cost.error(1, 1)
        elif cost_name == "rbf":
            cost.error(1, 1)
        else:
            cost.error(1, 2)


@pytest.mark.parametrize("cost_name", cost_names)
def test_costs_1D_noisy_names(signal_bkps_1D_noisy, cost_name):
    signal, bkps = signal_bkps_1D_noisy
    cost = cost_factory(cost_name)
    cost.fit(signal)
    cost.fit(signal.flatten())
    cost.error(0, 100)
    cost.error(100, signal.shape[0])
    cost.error(10, 50)
    cost.sum_of_costs(bkps)
    with pytest.raises(NotEnoughPoints):
        if cost_name == "cosine":
            cost.min_size = 4
            cost.error(1, 2)
        elif cost_name == "l2":
            cost.error(1, 1)
        elif cost_name == "rbf":
            cost.error(1, 1)
        else:
            cost.error(1, 2)


@pytest.mark.parametrize("cost_name", cost_names)
def test_costs_5D_names(signal_bkps_5D, cost_name):
    signal, bkps = signal_bkps_5D
    cost = cost_factory(cost_name)
    try:
        cost.fit(signal)
    except np.linalg.LinAlgError as e:
        if cost_name == "mahalanobis":
            reason_msg = (
                "The Mahalanobis metric matrix cannot be computed "
                "because the signal has a singular covariance matrix"
            )
            pytest.skip(reason_msg)
        else:
            raise
    cost.error(0, 100)
    cost.error(100, signal.shape[0])
    cost.error(10, 50)
    cost.sum_of_costs(bkps)
    if cost_name == "l2":
        assert cost.error(1, 2) == 0.0, "CostL2 on segment of size 1 returns 0."

    with pytest.raises(NotEnoughPoints):
        if cost_name == "cosine":
            cost.min_size = 4
            cost.error(1, 2)
        elif cost_name == "l2":
            cost.error(1, 1)
        elif cost_name == "rbf":
            cost.error(1, 1)
        else:
            cost.error(1, 2)


@pytest.mark.parametrize("cost_name", cost_names)
def test_costs_5D_noisy_names(signal_bkps_5D_noisy, cost_name):
    signal, bkps = signal_bkps_5D_noisy
    cost = cost_factory(cost_name)
    cost.fit(signal)
    cost.error(0, 100)
    cost.error(100, signal.shape[0])
    cost.error(10, 50)
    cost.sum_of_costs(bkps)
    with pytest.raises(NotEnoughPoints):
        if cost_name == "cosine":
            cost.min_size = 4
            cost.error(1, 2)
        elif cost_name == "l2":
            cost.error(1, 1)
        elif cost_name == "rbf":
            cost.error(1, 1)
        else:
            cost.error(1, 2)


def test_factory_exception():
    with pytest.raises(ValueError):
        cost_factory("Dummy cost name")


# Test CostLinear
def test_costlinear(signal_bkps_5D_noisy, signal_bkps_1D_noisy):
    # Creation of data. For convenience, we use
    # already generated signal_bkps_5D and signal_bkps_1D
    signal_regressors, _ = signal_bkps_5D_noisy  # regressors
    signal, bkps = signal_bkps_1D_noisy  # observed signal
    n = signal.shape[0]  # number of samples

    # First dimension is the observed signal.
    # Stack observed signal and regressors.
    # Add intercept to regressors
    s = np.c_[signal, signal_regressors, np.ones(n)]
    # Compute error
    c = CostLinear().fit(s)
    c.error(0, 100)
    c.error(100, n)
    c.error(10, 50)
    c.sum_of_costs(bkps)
    with pytest.raises(NotEnoughPoints):
        c.error(10, 11)


# Test CostNormal
def test_costnormal():
    # For signals that have truly constant segments, CostNormal should not fail
    # with the correction on the diagonal of the covariance matrix.

    # generate data
    signal_1D = np.r_[np.ones(100), np.arange(100), np.zeros(100)]
    signal_2D = np.c_[signal_1D, signal_1D[::-1]]
    n_samples = signal_1D.shape[0]
    bkps = [100, 200, n_samples]

    # test cost function
    for signal in (signal_1D, signal_2D):
        c = CostNormal(add_small_diag=True).fit(signal=signal)
        c.error(0, 100)
        c.error(100, n_samples)
        c.error(10, 50)
        c.sum_of_costs(bkps)
        with pytest.raises(NotEnoughPoints):
            c.error(10, 11)

    # test cost function without correction
    c = CostNormal(add_small_diag=False).fit(signal=signal_1D)
    assert np.isinf(c.error(0, 100))


def test_costml(signal_bkps_5D_noisy, signal_bkps_1D_noisy):
    """Test if `CostMl.fit` actually (re-)fits the metric matrix.

    Refitting the metric matrix should only happen if the user did not
    provide a metric matrix.
    """
    # no user-defined metric matrix
    c = CostMl()
    for signal, bkps in (signal_bkps_5D_noisy, signal_bkps_1D_noisy):
        c.fit(signal=signal)
        c.error(0, 100)
        c.sum_of_costs(bkps)
    # with a user-defined metric matrix
    signal, bkps = signal_bkps_5D_noisy
    _, n_dims = signal.shape
    c = CostMl(metric=np.eye(n_dims))
    c.fit(signal)
    c.error(10, 50)
    assert np.allclose(c.metric, np.eye(n_dims))


def test_costl2_small_data():
    """Test if CostL2 returns the correct segmentation for small data."""
    signal = np.array([0.0, 0.1, 1.2, 1.0])
    n_samples = signal.shape[0]
    algo = Binseg(model="l2", min_size=1, jump=1).fit(signal)
    computed_break_dict = {}
    for n_bkps in range(n_samples):
        try:
            result = algo.predict(n_bkps=n_bkps)
        except Exception as e:
            result = e
            computed_break_dict
        computed_break_dict[n_bkps] = result

    expected_break_dict = {0: [4], 1: [2, 4], 2: [2, 3, 4], 3: [1, 2, 3, 4]}
    err_msg = (
        f"Wrong segmentation, expected {expected_break_dict}, ",
        "got {computed_break_dict}.",
    )
    assert expected_break_dict == computed_break_dict, err_msg
